import matplotlib.pyplot as plt
import numpy as np
import rclpy
from rclpy.node import Node
from sensor_msgs.msg import LaserScan


class LidarToGridMap(Node):
    def _init_(self, grid_size=100, resolution=0.5):
        super()._init_("lidar_to_grid_map")
        self.grid_size = grid_size
        self.resolution = resolution
        self.grid = np.zeros((grid_size, grid_size))
        self.origin_x = grid_size // 2
        self.origin_y = grid_size // 2
        self.subscription = self.create_subscription(
            LaserScan, "/scan", self.lidar_callback, 10
        )
        self.subscription

    def lidar_callback(self, msg):
        angles = np.arange(msg.angle_min, msg.angle_max, msg.angle_increment)
        ranges = np.array(msg.ranges)

        plt.figure(figsize=(8, 8))
        plt.imshow(self.grid, cmap="gray", extent=[-25, 25, -25, 25], origin="lower")

        for r, angle in zip(ranges, angles):
            if msg.range_min <= r <= msg.range_max:
                x = int(self.origin_x + (r * np.cos(angle)) / self.resolution)
                y = int(self.origin_y + (r * np.sin(angle)) / self.resolution)
                if 0 <= x < self.grid_size and 0 <= y < self.grid_size:
                    self.grid[y, x] = 1
                    plt.plot(
                        [
                            self.origin_x * self.resolution - 25,
                            x * self.resolution - 25,
                        ],
                        [
                            self.origin_y * self.resolution - 25,
                            y * self.resolution - 25,
                        ],
                        "g-",
                    )

        plt.title("Grid Map de Ocupação com Linhas do LIDAR")
        plt.xlabel("X (metros)")
        plt.ylabel("Y (metros)")
        plt.grid(True)
        plt.show()


def main(args=None):
    rclpy.init(args=args)
    node = LidarToGridMap()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()



import functools
import os

import cv2
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.pyplot import imread
from yaml import SafeLoader, load


class Map2D:
    def __init__(
        self,
        folder=None,
        name=None,
        silent=False,
    ):
        self.path = folder

        if folder is None or name is None:
            return

        folder = os.path.expanduser(folder)
        yaml_file = os.path.join(folder, name + ".yaml")

        if not silent:
            print(f"Loading map definition from {yaml_file}")

        with open(yaml_file) as stream:
            mapparams = load(stream, Loader=SafeLoader)
        map_file = os.path.join(folder, mapparams["image"])

        if not silent:
            print(f"Map definition found. Loading map from {map_file}")

        mapimage = imread(map_file)
        temp = (1.0 - mapimage.T[:, ::-1] / 254.0).astype(np.float32)
        mapimage = np.ascontiguousarray(temp)
        self._occupancy = mapimage
        self.occupancy_shape0 = mapimage.shape[0]
        self.occupancy_shape1 = mapimage.shape[1]
        self.resolution_ = mapparams["resolution"]
        self.origin = np.array(mapparams["origin"][:2]).astype(np.float32)

        if mapparams["origin"][2] != 0:
            raise ValueError("Map origin z coordinate must be 0")

        self._thresh_occupied = mapparams["occupied_thresh"]
        self.thresh_free = mapparams["free_thresh"]
        self.HUGE_ = 100 * self.occupancy_shape0 * self.occupancy_shape1

        if self.resolution_ == 0:
            raise ValueError("resolution can not be 0")

    def occupancy_grid(self) -> np.ndarray:
        """return the gridmap without filter

        Returns:
            np.ndarray: occupancy grid
        """
        occ = np.array(self._occupancy)
        return occ

    @functools.lru_cache(maxsize=None)
    def _grid_map(self) -> np.ndarray:
        """This function receives the grid map and filters only the region of the map

        Returns:
            np.ndarray: grid map
        """
        data = self.occupancy_grid()

        data = np.where(data < 0, 0, data)
        data = np.where(data != 0, 1, data)

        idx = np.where(data == 0)

        min_x = np.min(idx[1])
        max_x = np.max(idx[1])
        min_y = np.min(idx[0])
        max_y = np.max(idx[0])

        dist_x = (max_x - min_x) + 1
        dist_y = (max_y - min_y) + 1

        if (max_y - min_y) != (max_x - min_x):
            dist_y = max_y - min_y
            dist_x = max_x - min_x

            diff = round(abs(dist_y - dist_x) / 2)

            # distance y > distance x
            if dist_y > dist_x:
                min_x = int(min_x - diff)
                max_x = int(max_x + diff)

            # distance y < distance x
            if dist_y < dist_x:
                min_y = int(min_y - diff)
                max_y = int(max_y + diff)

        diff_x = max_x - min_x
        diff_y = max_y - min_y

        # TODO: remove this
        if abs((diff_y) - (diff_x)) == 1:

            if diff_y < diff_x:
                max_y = max_y + 1

            if diff_y > diff_x:
                max_x = max_x + 1

        if min(min_x, max_x, min_y, max_y) < 0:
            min_x_adjusted = min_x + abs(min_x)
            max_x_adjusted = max_x + abs(min_x)
            min_y_adjusted = min_y + abs(min_y)
            max_y_adjusted = max_y + abs(min_y)

            map_record = data[
                min_y_adjusted : max_y_adjusted + 1, min_x_adjusted : max_x_adjusted + 1
            ]

        else:
            map_record = data[min_y : max_y + 1, min_x : max_x + 1]

        new_map_grid = np.zeros_like(map_record)
        new_map_grid[map_record == 0] = 1

        return new_map_grid

    def initial_environment2d(
        self,
        plot=True,
        kernel_size=(3, 3),
        morph_iterations=1,
        approx_epsilon_factor=0.030,
        contour_retrieval_mode=cv2.RETR_TREE,
        contour_approx_method=cv2.CHAIN_APPROX_SIMPLE,
    ):

        new_map_grid = self._grid_map()

        print(f"Shape do new_map_grid: {new_map_grid.shape}")

        idx = np.where(new_map_grid.sum(axis=0) > 0)[0]

        if idx.size == 0:
            return

        min_idx = np.min(idx)
        max_idx = np.max(idx)

        subgrid = new_map_grid[:, min_idx : max_idx + 1]

        subgrid_uint8 = (subgrid * 255).astype(np.uint8)

        kernel = np.ones(kernel_size, np.uint8)

        eroded = cv2.erode(subgrid_uint8, kernel, iterations=morph_iterations)
        dilated = cv2.dilate(eroded, kernel, iterations=morph_iterations)

        num_labels, labels, stats, centroids = cv2.connectedComponentsWithStats(
            dilated, connectivity=8
        )

        if num_labels <= 1:
            return

        largest_component = 1 + np.argmax(stats[1:, cv2.CC_STAT_AREA])

        mask = np.zeros_like(dilated)
        mask[labels == largest_component] = 255

        mask_closed = cv2.morphologyEx(
            mask, cv2.MORPH_CLOSE, kernel, iterations=morph_iterations
        )

        contours, hierarchy = cv2.findContours(
            mask_closed, contour_retrieval_mode, contour_approx_method
        )

        if not contours:
            return

        contour_mask = np.zeros_like(mask_closed)

        for i, contour in enumerate(contours):
            epsilon = approx_epsilon_factor * cv2.arcLength(contour, True)
            approx = cv2.approxPolyDP(contour, epsilon, True)

            if hierarchy[0][i][3] == -1:
                cv2.fillPoly(contour_mask, [approx], 255)
            else:
                cv2.fillPoly(contour_mask, [approx], 0)

        kernel_smooth = np.ones((1, 1), np.uint8)
        contour_mask = cv2.morphologyEx(contour_mask, cv2.MORPH_OPEN, kernel_smooth)
        print(contour_mask)

        if plot:

            plt.figure(figsize=(6, 6))
            plt.imshow(contour_mask, cmap="gray")
            plt.axis("off")
            plt.show()

        return contour_mask

    def plot_initial_environment3d(self, plot=True) -> None:
        """generate environment from map"""

        self.initial_environment2d(plot=False)

    #     # new_map_grid = self._grid_map()

    #     idx = np.where(map_grid.sum(axis=0) > 0)[0]

    #     min_idx = int(np.min(idx))
    #     max_idx = int(np.max(idx))

    # print(new_map_grid)

    # all_edges = []

    # for i in tqdm(range(min_idx, max_idx), desc="Plotting environment"):
    #     for j in range(min_idx, max_idx):
    #         if new_map_grid[i, j] == 1:
    #             polygon = [(j, i), (j + 1, i), (j + 1, i + 1), (j, i + 1)]
    #             poly = Polygon(polygon, color=(0.1, 0.2, 0.5, 0.15))

    #             vert = poly.get_xy()
    #             edges = [
    #                 (vert[k], vert[(k + 1) % len(vert)]) for k in range(len(vert))
    #             ]

    #             all_edges.extend(edges)



from typing import Tuple

import numpy as np
from numba import njit


@njit
def distance_to_goal(x: float, y: float, goal_x: float, goal_y: float) -> float:
    return np.sqrt((x - goal_x) ** 2 + (y - goal_y) ** 2)


def angle_to_goal(
    x: float, y: float, theta: float, goal_x: float, goal_y: float
) -> float:
    o_t = np.array([np.cos(theta), np.sin(theta)])
    g_t = np.array([goal_x - x, goal_y - y])

    # Ângulo alpha
    cross_product = np.cross(o_t, g_t)
    dot_product = np.dot(o_t, g_t)

    alpha = np.abs(np.arctan2(np.linalg.norm(cross_product), dot_product))

    # Cálculo de alpha_norm
    if alpha <= 0.0:
        alpha = 0.1
    elif alpha >= 3.2:
        alpha = 3.3

    return alpha


def min_laser(measurement: np.ndarray, threshold: float = 0.1):
    laser = np.min(measurement)
    if laser < threshold:
        return True, laser
    else:
        return False, laser


def get_reward(
    measurement, distance, collision, alpha, distance_init
) -> Tuple[np.float64, np.bool_]:

    reward = 0.0

    alpha_norm = (1 / alpha) * 0.2

    if alpha_norm >= 0.0 and alpha_norm <= 0.7:
        alpha_norm = 2
    else:
        alpha_norm = 0

    if distance < 0.4:
        return 500.0, np.bool_(True)
    elif collision:
        return -500.0, np.bool_(True)
    else:
        obstacle = r3(min(measurement))
        if distance < 0.4:
            obstacle = 0

        reward = distance_init + abs(alpha_norm) - abs(distance) - obstacle
        return np.float32(reward), np.bool_(False)


@njit
def r3(x):
    if x < 0.3:
        return 1 - x
    else:
        return 0.0


@njit
def normalize(value, min_value, max_value):
    return (value - min_value) / (max_value - min_value)



from functools import lru_cache
from typing import List, Tuple

import numpy as np
from numba import njit

RED_BOLD = "\033[1;31m"
RESET = "\033[0m"


class Collision:

    def check_collision(
        self, exterior: list, interiors: list, xp: np.ndarray, yp: np.ndarray
    ) -> bool:
        try:
            return is_circle_in_polygon(exterior, interiors, xp, yp)
        except Exception as e:
            print(f"{RED_BOLD}Error in check_collision: {e}{RESET}")
            raise

    def filter_segments(self, segs: List, x: float, y: float, max_range: int):
        try:
            return filter_list_segment(segs, x, y, max_range)
        except Exception as e:
            print(f"{RED_BOLD}Error in filter_segment: {e}{RESET}")
            raise

    def extract_seg_from_polygon(self, stack: List) -> List:
        try:
            return extract_segment_from_polygon(stack)
        except Exception as e:
            print(f"{RED_BOLD}Error in extract_segment_from_polygon: {e}{RESET}")
            raise

    def lidar_intersection(
        self,
        robot_x: float,
        robot_y: float,
        lidar_range: float,
        lidar_angles: np.ndarray,
        segments: List,
    ):
        try:
            return lidar_intersections(
                robot_x, robot_y, lidar_range, lidar_angles, segments
            )
        except Exception as e:
            print(f"{RED_BOLD}Error in lidar_intersection: {e}{RESET}")
            raise

    def lidar_measurement(
        self,
        robot_x: float,
        robot_y: float,
        lidar_range: float,
        lidar_angles: np.ndarray,
        segments: List,
    ):
        try:
            return lidar_measurements(
                robot_x, robot_y, lidar_range, lidar_angles, segments
            )
        except Exception as e:
            print(f"{RED_BOLD}Error in lidar_measurement: {e}{RESET}")
            raise


@njit
def filter_list_segment(segs: List, x: float, y: float, max_range: int) -> List:
    """
    Filters line segments based on proximity to a given point.

    Parameters:
    segs (list): List of line segments.
    x (float): X-coordinate of the point.
    y (float): Y-coordinate of the point.
    max_range (int): Maximum range for filtering segments.

    Returns:
    List: List of line segments filtered based on proximity to the given point.
    """
    segments_trans = [
        [np.array(segments[:2]), np.array(segments[2:])] for segments in segs
    ]

    segments_inside = []
    region_center = np.array([x, y])

    for segment in segments_trans:
        x1, y1 = segment[0]
        x2, y2 = segment[1]
        segment_endpoints = np.array([[x1, y1], [x2, y2]])

        distances = calculate_distances(segment_endpoints, region_center)

        if is_distance_within_range(distances, max_range):
            segments_inside.append(segment)

    return segments_inside


@njit
def is_distance_within_range(distance: np.ndarray, lidar_range: int) -> bool:
    """
    Determines if a distance is within the range of the lidar sensor.

    Parameters:
    distance (float): The distance to be checked.
    lidar_range (int): The maximum range of the lidar sensor.

    Returns:
    bool: True if the distance is within the range, False otherwise.
    """

    return np.all(distance <= lidar_range)


@njit
def calculate_distances(seg_ends: np.ndarray, center: np.ndarray) -> np.ndarray:
    """
    Calculates the distances from the endpoints of line segments to a given central point.

    Parameters:
    segment_endpoints (np.ndarray): Array of line segment endpoints. Each row represents a point.
    region_center (np.ndarray): Coordinates of the central point.

    Returns:
    np.ndarray: Array of distances from each endpoint to the central point.
    """

    distances = np.sqrt(np.sum((seg_ends - center) ** 2, axis=1))

    return distances


@njit
def extract_segment_from_polygon(stack: List) -> List:
    """
    Extracts line segments from a stack of polygons.

    Parameters:
    stack (List): List of polygons, each represented by a list of points.

    Returns:
    List: List of line segments extracted from the polygons.
    """

    assert len(stack) >= 1, "Stack must have at least 1 polygon."

    total_segments = []
    for polygon in stack:
        segments = connect_polygon_points(polygon)
        total_segments.extend(segments)

    return convert_to_line_segments(total_segments)


@njit
def convert_to_line_segments(total_segments: List) -> List:
    """
    Converts a list of line segments into formed (x1, y1, x2, y2) representing start point, end point

    Parameters:
    total_segments (List): List of line segments.

    Returns:
    List: List of line segments converted into (x1, y1, x2, y2) format.
    """

    assert len(total_segments) >= 3, "Polygon must have at least 3 points."

    line_segments = []
    for segment in total_segments:
        line_segments.append(
            (segment[0][0], segment[0][1], segment[1][0], segment[1][1])
        )

    return line_segments


@njit
def connect_polygon_points(polygon: np.ndarray) -> List:
    """
    Connects the points of a polygon to form line segments.

    Parameters:
    polygon (np.ndarray): The polygon represented by a list of points.

    Returns:
    List: List of line segments connecting the points of the polygon.
    """

    #                segment 1
    # (point 1) +-----------------+ (point 2)
    #           |                 |
    # segment 4 |                 | segment 2
    #           |                 |
    # (point 4) +-----------------+ (point 3)
    #                segment 3

    assert len(polygon) >= 3, "Polygon must have at least 3 points."

    num_points = len(polygon)
    segments = []
    for i in range(num_points):
        current_point = polygon[i]
        # wrap-around to close the polygon
        next_point = polygon[(i + 1) % num_points]
        segment = (current_point, next_point)
        segments.append(segment)

    return segments


lru_cache(maxsize=5)


def intersect_segment_with_polygon(segment: list, poly: list) -> List:
    """
    Intersects a line segment with a polygon.

    Parameters:
    segment (list): Line segment represented by two points.
    poly (list): List of line segments representing a polygon.

    Returns:
    List: List of intersection points with p, t, r
    """

    # (q+s)  (p+r)
    #   *     *
    #    \   /
    #     \ /
    #      *  temp
    #     / \
    #    /   \
    #   /     \
    #  *       *
    # (p)      (q)

    inter = []

    for segments in poly:
        p = segment[0]
        r = segment[1] - segment[0]
        q = segments[0]
        s = segments[1] - segments[0]

        temp = cross_product_2d_vector(r, s)

        # If temp is 0, the lines are parallel (//)
        if temp != 0:
            direction_vec = q - p

            t, u = calculate_temp_vector(direction_vec, r, s)

            # If t and u are between 0 and 1
            if t >= 0 and t <= 1 and u >= 0 and u <= 1:
                interception_point = calculate_intersection_point_vec(p, t, r)
                distances = calculate_distance_to_intersection(interception_point, p)
                inter.append([True, interception_point, distances])

    return inter


@njit
def calculate_intersection_point_vec(p: float, t: float, r: float) -> float:
    """
    Calculates the intersection point of two line segments.

    Parameters:
    p (float): The starting point.
    t (float): The t value.
    r (float): The direction vector.

    Returns:
    float: The intersection point.
    """
    return p + t * r


@njit
def calculate_distance_to_intersection(inter_point: np.ndarray, p: np.ndarray) -> float:
    """
    Calculates the distance to the intersection point.

    Parameters:
    inter_point (np.ndarray): The intersection point.
    p (np.ndarray): The starting point.

    Returns:
    float: The distance to the intersection point.
    """
    return np.sqrt(np.sum((inter_point - p) ** 2))


@njit
def calculate_temp_vector(
    direction_vec: np.ndarray, r: np.ndarray, s: np.ndarray
) -> Tuple[float, float]:
    """
    Calculates the t and u values for the intersection of two line segments.

    Parameters:
    vector_direction (np.ndarray): The direction vector.
    vector_r (np.ndarray): The vector r.
    vector_s (np.ndarray): The vector s.

    Returns:
    Tuple[float, float]: The t and u values.
    """
    cross_product = cross_product_2d_vector(r, s)

    cross_product_direction_vec_s = cross_product_2d_vector(direction_vec, s)
    cross_product_direction_vec_r = cross_product_2d_vector(direction_vec, r)

    # If cross_product is not 0, calculate t and u
    if cross_product != 0:
        t = cross_product_direction_vec_s / cross_product
        u = cross_product_direction_vec_r / cross_product
    else:
        t = u = 1

    return t, u


# TODO: refactor
lru_cache(maxsize=5)


def lidar_intersections(
    robot_x: float,
    robot_y: float,
    lidar_range: float,
    lidar_angles: np.ndarray,
    segments: List,
):

    intersections = []
    for i, angle in enumerate(lidar_angles):
        lidar_segment = lidar_to_segment(robot_x, robot_y, lidar_range, angle)

        lidar_segment_transformed = [np.array(segmento) for segmento in lidar_segment]

        intersected = intersect_segment_with_polygon(
            lidar_segment_transformed, segments
        )

        if intersected:
            intercept, position = position_intersection(intersected)

            if intercept:
                intersections.append(position)

            else:
                intersections.append((0.0, 0.0))

        else:
            intersections.append((0.0, 0.0))
            continue

    return intersections


# TODO: Add (njit, docstring)
def position_intersection(
    intersection: List[Tuple[bool, np.ndarray, float]]
) -> Tuple[bool, Tuple]:

    min_lrange_index = np.argmin([point[2] for point in intersection])
    inter_point = intersection[min_lrange_index][1]
    inter_point_rounded = [round(coord, 3) for coord in inter_point]

    return True, tuple(inter_point_rounded)


# TODO: refactor
def lidar_measurements(
    robot_x: float,
    robot_y: float,
    lidar_range: float,
    lidar_angles: np.ndarray,
    segments: List,
):

    measurements = []
    for i, angle in enumerate(lidar_angles):
        lidar_segment = lidar_to_segment(robot_x, robot_y, lidar_range, angle)
        lidar_segment_transformed = [np.array(segmento) for segmento in lidar_segment]

        intersected = intersect_segment_with_polygon(
            lidar_segment_transformed, segments
        )

        if intersected:
            intercept, ranges = measurements_intersection(intersected)

            if intercept:
                measurements.append(ranges)

            else:
                measurements.append(0.2)
        else:
            measurements.append(0.2)
            continue

    return measurements


# TODO: Add (njit, docstring)
def measurements_intersection(intersection: List) -> Tuple[bool, float]:

    min_lrange_index = np.argmin([point[2] for point in intersection])
    lrange = round(intersection[min_lrange_index][2], 3)

    return True, lrange


@njit
def lidar_to_segment(
    robot_x: float, robot_y: float, lidar_range: float, angle: float
) -> List[Tuple[float, float]]:
    return [
        (robot_x, robot_y),
        (robot_x + lidar_range * np.cos(angle), robot_y + lidar_range * np.sin(angle)),
    ]


@njit(fastmath=True)
def cross_product_2d_vector(a: np.ndarray, b: np.ndarray) -> float:
    """
    Calculates the cross product of two 2D vectors.

    Parameters:
    a (np.ndarray): The first 2D vector.
    b (np.ndarray): The second 2D vector.

    Returns:
    float: The cross product of the two vectors.
    """

    assert a.shape == b.shape, "Vectors must have the same shape in 2D cross product."
    assert a.shape[0] == 2, "Vectors must be 2D for cross product."
    return a[0] * b[1] - a[1] * b[0]


@njit
def ray_casting(edges: list, xp: np.ndarray, yp: np.ndarray) -> bool:
    """
    Ray casting algorithm to check if a point is inside a polygon.

    Parameters:
    edges (list): The edges of the polygon.
    xp (np.ndarray): The x coordinates of the point.
    yp (np.ndarray): The y coordinates of the point.

    Returns:
    bool: True if the point is inside the polygon, False otherwise
    """

    cnt = 0
    for edge in edges:
        (x1, y1, x2, y2) = edge
        if (yp < y1) != (yp < y2) and xp < x1 + ((yp - y1) / (y2 - y1)) * (x2 - x1):
            cnt += 1
    return cnt % 2 == 1


def is_point_in_polygon(exterior: list, interiors: list, xp: float, yp: float) -> bool:

    if not ray_casting(exterior, xp, yp):
        return False

    for hole in interiors:
        if ray_casting(hole, xp, yp):
            return False

    return True


def is_circle_in_polygon(
    exterior: list,
    interiors: list,
    xc: float,
    yc: float,
    radius: float = 0.3,
    num_segments: int = 100,
) -> bool:
    """
    Checks if a circle is completely within a polygon (including avoiding holes).

    Parameters:
    - exterior: List of edges defining the outer boundary of the polygon.
    - interiors: List of lists of edges defining holes in the polygon.
    - xc: X coordinate of the circle's center.
    - yc: Y coordinate of the circle's center.
    - radius: Radius of the circle.
    - num_segments: Number of segments to approximate the circle boundary.

    Returns:
    - bool: True if the circle is completely inside the polygon and not in any hole, False otherwise.
    """
    angles = np.linspace(0, 2 * np.pi, num_segments)
    for angle in angles:
        xp = xc + radius * np.cos(angle)
        yp = yc + radius * np.sin(angle)
        if not is_point_in_polygon(exterior, interiors, xp, yp):
            return False
    return True



# from gymnasium.envs.registration import register

# register(
#     id="microvault/NaviEnv-v0",
#     entry_point="microvault.environment.environment_navigation:NaviEnv",
# )

from rnl.training.interface import Trainer, make, render, robot, sensor

__all__ = ["robot", "sensor", "render", "make", "Trainer"]



import gymnasium as gym
import matplotlib.pyplot as plt
import numpy as np

from rnl.algorithms.rainbow import RainbowDQN
from rnl.configs.config import AgentConfig, TrainerConfig


def make_vect_envs(env_name, num_envs=1, **env_kwargs):
    """Returns async-vectorized gym environments with custom parameters.

    :param env_name: Gym environment name or custom environment class
    :type env_name: str or type
    :param num_envs: Number of vectorized environments, defaults to 1
    :type num_envs: int, optional
    :param env_kwargs: Additional keyword arguments for the environment
    :type env_kwargs: dict
    """

    def make_env():
        if isinstance(env_name, str):
            return gym.make(env_name, **env_kwargs)
        else:
            return env_name(**env_kwargs)

    return gym.vector.AsyncVectorEnv([make_env for _ in range(num_envs)])


def make_skill_vect_envs(env_name, skill, num_envs=1):
    """Returns async-vectorized gym environments.

    :param env_name: Gym environment name
    :type env_name: str
    :param skill: Skill wrapper to apply to environment
    :type skill: agilerl.wrappers.learning.Skill
    :param num_envs: Number of vectorized environments, defaults to 1
    :type num_envs: int, optional
    """
    return gym.vector.AsyncVectorEnv(
        [lambda: skill(gym.make(env_name)) for i in range(num_envs)]
    )


def create_population(
    algo,
    state_dim,
    action_dim,
    one_hot,
    net_config,
    agent_config: AgentConfig,
    trainer_config: TrainerConfig,
    actor_network=None,
    critic_network=None,
    population_size=1,
    num_envs=1,
    device="cpu",
    accelerator=None,
):
    """Returns population of identical agents.

    :param algo: RL algorithm
    :type algo: str
    :param state_dim: State observation dimension
    :type state_dim: int
    :param action_dim: Action dimension
    :type action_dim: int
    :param one_hot: One-hot encoding
    :type one_hot: bool
    :param INIT_HP: Initial hyperparameters
    :type INIT_HP: dict
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: nn.Module, optional
    :param critic_network: Custom critic network, defaults to None
    :type critic_network: nn.Module, optional
    :param population_size: Number of agents in population, defaults to 1
    :type population_size: int, optional
    :param num_envs: Number of vectorized environments, defaults to 1
    :type num_envs: int, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    """
    population = []

    for idx in range(population_size):
        agent = RainbowDQN(
            state_dim=state_dim,
            action_dim=action_dim,
            one_hot=one_hot,
            index=idx,
            net_config=net_config,
            batch_size=trainer_config.batch_size,
            lr=trainer_config.lr,
            learn_step=trainer_config.learn_step,
            gamma=agent_config.gamma,
            tau=agent_config.tau,
            beta=agent_config.beta,
            prior_eps=agent_config.prior_eps,
            num_atoms=agent_config.num_atoms,
            v_min=agent_config.v_min,
            v_max=agent_config.v_max,
            n_step=agent_config.n_step,
            actor_network=actor_network,
            device=device,
            accelerator=accelerator,
        )
        population.append(agent)

    return population


def calculate_vectorized_scores(
    rewards, terminations, include_unterminated=False, only_first_episode=True
):
    episode_rewards = []
    num_envs, _ = rewards.shape

    for env_index in range(num_envs):
        # Find the indices where episodes terminate for the current environment
        termination_indices = np.where(terminations[env_index] == 1)[0]

        # If no terminations, sum the entire reward array for this environment
        if len(termination_indices) == 0:
            episode_reward = np.sum(rewards[env_index])
            episode_rewards.append(episode_reward)
            continue  # Skip to the next environment

        # Initialize the starting index for segmenting
        start_index = 0

        for termination_index in termination_indices:
            # Sum the rewards for the current episode
            episode_reward = np.sum(
                rewards[env_index, start_index : termination_index + 1]
            )

            # Store the episode reward
            episode_rewards.append(episode_reward)

            # If only the first episode is required, break after processing it
            if only_first_episode:
                break

            # Update the starting index for segmenting
            start_index = termination_index + 1

        # If include_unterminated is True, sum the rewards from the last termination index to the end
        if (
            not only_first_episode
            and include_unterminated
            and start_index < len(rewards[env_index])
        ):
            episode_reward = np.sum(rewards[env_index, start_index:])
            episode_rewards.append(episode_reward)

    return episode_rewards


def print_hyperparams(pop):
    """Prints current hyperparameters of agents in a population and their fitnesses.

    :param pop: Population of agents
    :type pop: list[object]
    """

    for agent in pop:
        print(
            "Agent ID: {}    Mean 5 Fitness: {:.2f}    Attributes: {}".format(
                agent.index, np.mean(agent.fitness[-5:]), agent.inspect_attributes()
            )
        )


def plot_population_score(pop):
    """Plots the fitness scores of agents in a population.

    :param pop: Population of agents
    :type pop: list[object]
    """
    plt.figure()
    for agent in pop:
        scores = agent.fitness
        steps = agent.steps[:-1]
        plt.plot(steps, scores)
    plt.title("Score History - Mutations")
    plt.xlabel("Steps")
    plt.ylim(bottom=-400)
    plt.show()



import numpy as np

# import numpy.random as npr
from numba import njit


class GenerateWorld:
    def __init__(self):
        pass

    def generate_maze(
        self,
        map_size: int,
        decimation: float = 0.0,
        min_blocks: int = 10,
        num_cells_togo: int = 100,
    ) -> np.ndarray:
        return generate_maze(map_size, decimation, min_blocks, num_cells_togo)


# @njit(parallel=True)
# def generate_maze(
#     map_size: int,
#     decimation: float = 0.0,
#     min_blocks: int = 10,
#     num_cells_togo: int = 150,
# ) -> np.ndarray:
#     """
#     Generates a maze using Kruskal's algorithm.

#     Parameters:
#     map_size (int): The size of the maze.
#     decimation (float): The probability of removing blocks.
#     min_blocks (int): The minimum number of blocks to keep.
#     num_cells_togo (int): The number of cells to remove.

#     Returns:
#     np.ndarray: The generated maze represented as a binary array.
#     """
#     m = (map_size - 1) // 2
#     n = (map_size - 1) // 2
#     maze = np.ones((map_size, map_size))
#     maze = np.ones((map_size, map_size))
#     for i in range(m):
#         for j in range(n):
#             maze[2 * i + 1, 2 * j + 1] = 0
#     m = m - 1
#     L = np.arange(n + 1)
#     R = np.arange(n)
#     L[n] = n - 1

#     while m > 0:
#         for i in range(n):
#             j = L[i + 1]
#             if i != j and npr.randint(3) != 0:
#                 R[j] = R[i]
#                 L[R[j]] = j
#                 R[i] = i + 1
#                 L[R[i]] = i
#                 maze[2 * (n - m) - 1, 2 * i + 2] = 0
#             if i != L[i] and npr.randint(3) != 0:
#                 L[R[i]] = L[i]
#                 R[L[i]] = R[i]
#                 L[i] = i
#                 R[i] = i
#             else:
#                 maze[2 * (n - m), 2 * i + 1] = 0

#         m -= 1

#     for i in range(n):
#         j = L[i + 1]
#         if i != j and (i == L[i] or npr.randint(3) != 0):
#             R[j] = R[i]
#             L[R[j]] = j
#             R[i] = i + 1
#             L[R[i]] = i
#             maze[2 * (n - m) - 1, 2 * i + 2] = 0

#         L[R[i]] = L[i]
#         R[L[i]] = R[i]
#         L[i] = i
#         R[i] = i

#     # ----- Generate Map -----
#     index_ones = np.arange(map_size * map_size)[maze.flatten() == 1]

#     if index_ones.size < min_blocks:
#         raise ValueError("Minimum number of blocks cannot be placed.")

#     reserve = min(index_ones.size, min_blocks)
#     num_cells_togo = min(num_cells_togo, index_ones.size - reserve)

#     for _ in range(num_cells_togo):
#         blocks_remove = np.random.randint(0, map_size**2)
#         row_index = blocks_remove // map_size
#         col_index = blocks_remove % map_size
#         maze[row_index, col_index] = 0

#     return maze


@njit(parallel=True)
def generate_maze(
    map_size: int,
    decimation: float = 0.0,
    min_blocks: int = 1,
    num_cells_togo: int = 150,
) -> np.ndarray:
    """
    Generates a maze using Kruskal's algorithm without randomness.

    Parameters:
    map_size (int): The size of the maze.
    decimation (float): The probability of removing blocks.
    min_blocks (int): The minimum number of blocks to keep.
    num_cells_togo (int): The number of cells to remove.

    Returns:
    np.ndarray: The generated maze represented as a binary array.
    """
    m = (map_size - 1) // 2
    n = (map_size - 1) // 2
    maze = np.ones((map_size, map_size))
    for i in range(m):
        for j in range(n):
            maze[2 * i + 1, 2 * j + 1] = 0

    m = m - 1
    L = np.arange(n + 1)
    R = np.arange(n)
    L[n] = n - 1

    while m > 0:
        for i in range(n):
            j = L[i + 1]
            if i != j and (i + j) % 2 == 0:
                R[j] = R[i]
                L[R[j]] = j
                R[i] = i + 1
                L[R[i]] = i
                maze[2 * (n - m) - 1, 2 * i + 2] = 0

            if i != L[i] and (i + j) % 2 == 0:
                L[R[i]] = L[i]
                R[L[i]] = R[i]
                L[i] = i
                R[i] = i
            else:
                maze[2 * (n - m), 2 * i + 1] = 0

        m -= 1

    for i in range(n):
        j = L[i + 1]
        if i != j and ((i + j) % 2 == 0):
            R[j] = R[i]
            L[R[j]] = j
            R[i] = i + 1
            L[R[i]] = i
            maze[2 * (n - m) - 1, 2 * i + 2] = 0

        L[R[i]] = L[i]
        R[L[i]] = R[i]
        L[i] = i
        R[i] = i

    # ----- Generate Map -----
    index_ones = np.arange(map_size * map_size)[maze.flatten() == 1]

    if index_ones.size < min_blocks:
        raise ValueError("Minimum number of blocks cannot be placed.")

    reserve = min(index_ones.size, min_blocks)
    num_cells_togo = min(num_cells_togo, index_ones.size - reserve)

    for i in range(num_cells_togo):
        blocks_remove = (i * 7) % (map_size**2)
        row_index = blocks_remove // map_size
        col_index = blocks_remove % map_size
        maze[row_index, col_index] = 0

    return maze



import warnings
from dataclasses import dataclass
from typing import Union

import torch


@dataclass
class OutputSpec:
    module_name: str = None
    range: Union[list, tuple] = None
    negate: bool = False
    check_nan: bool = False
    check_inf: bool = False

    @property
    def name(self):
        if self.module_name is None:
            return "Module's output"
        else:
            return f"Module {self.module_name}'s output"

    @property
    def condition(self):
        low, high = self.range
        if low is None:
            return f"< {high}"
        elif high is None:
            return f"> {low}"
        else:
            return f"> {low} and < {high}"

    def update(
        self,
        module_name=None,
        range=None,
        negate=False,
        check_nan=False,
        check_inf=False,
    ):
        if module_name is not None and module_name != self.module_name:
            old_name = self.name
            self.module_name = module_name
            warnings.warn(f"{old_name} is renamed as {self.name}.")
        if range is not None:
            self.range = range
            self.negate = negate
        if check_nan:
            self.check_nan = True
        if check_inf:
            self.check_inf = True

    def validate(self, output):
        error_items = []
        if self.range is not None:
            error_items.append(self.validate_range(output))
        if self.check_nan:
            error_items.append(self.validate_nan(output))
        if self.check_inf:
            error_items.append(self.validate_inf(output))

        error_items = [_ for _ in error_items if _ is not None]
        if len(error_items):
            raise RuntimeError("validate")

    def validate_range(self, output):
        low, high = self.range
        status = torch.ones_like(output, dtype=torch.bool)
        if low is not None:
            status = output >= low
        if high is not None:
            status = status & (output <= high)

        if not self.negate:
            if not torch.all(status).item():
                return (
                    f"{self.name} should all {self.condition}. " "Some are out of range"
                )
        else:
            if torch.all(status).item():
                return f"{self.name} shouldn't all {self.condition}"

    def validate_nan(self, output):
        if torch.any(torch.isnan(output)).item():
            return f"{self.name} contains NaN."

    def validate_inf(self, output):
        if torch.any(torch.isinf(output)).item():
            return f"{self.name} contains inf."



from typing import List

from rnl.configs.config import (
    AgentConfig,
    EnvConfig,
    HPOConfig,
    NetworkConfig,
    RenderConfig,
    RobotConfig,
    SensorConfig,
    TrainerConfig,
)
from rnl.training.learn import inference, training


def robot(
    base_radius: float,
    vel_linear: List,
    vel_angular: List,
    wheel_distance: float,
    weight: float,
    threshold: float,
):
    return RobotConfig(
        base_radius, vel_linear, vel_angular, wheel_distance, weight, threshold
    )


def sensor(fov: float, num_rays: int, min_range: float, max_range: float):
    return SensorConfig(fov, num_rays, min_range, max_range)


def make(
    map_file: str,
    random_mode: str,
    timestep: int,
    grid_dimension: int,
    friction: float,
    porcentage_obstacles: float,
    max_step: int = 1000,
):
    return EnvConfig(
        map_file,
        random_mode,
        timestep,
        grid_dimension,
        friction,
        porcentage_obstacles,
        max_step,
    )


def render(fps: int, controller: bool, rgb_array: bool):
    return RenderConfig(fps, controller, rgb_array)


class Trainer:
    def __init__(
        self,
        robot_config: RobotConfig = RobotConfig(),
        sensor_config: SensorConfig = SensorConfig(),
        env_config: EnvConfig = EnvConfig(),
        render_config: RenderConfig = RenderConfig(),
        pretrained_model=False,
    ):
        self.robot_config = robot_config
        self.sensor_config = sensor_config
        self.env_config = env_config
        self.render_config = render_config
        self.pretrained_model = pretrained_model

    def learn(
        self,
        max_timestep: int = 800000,
        memory_size: int = 1000000,
        gamma: float = 0.99,
        n_step: int = 3,
        alpha: float = 0.6,
        beta: float = 0.4,
        tau: float = 0.001,
        prior_eps: float = 0.000001,
        num_atoms: int = 51,
        v_min: int = -200,
        v_max: int = 200,
        epsilon_start: float = 1.0,
        epsilon_end: float = 0.1,
        epsilon_decay: float = 0.995,
        batch_size: int = 64,  # Batch size
        lr: float = 0.0001,  # Learning rate
        seed: int = 1,
        num_envs: int = 16,
        device: str = "mps",
        learn_step: int = 10,
        target_score: int = 200,
        max_steps: int = 1000000,
        evaluation_steps: int = 10000,
        evaluation_loop: int = 1,
        learning_delay: int = 0,
        n_step_memory: int = 1,
        checkpoint: int = 100,
        checkpoint_path: str = "checkpoints",
        overwrite_checkpoints: bool = False,
        use_wandb: bool = False,
        wandb_api_key: str = "",
        accelerator: bool = False,
        use_mutation: bool = True,
        freq_evolution: int = 10000,
        population_size: int = 6,
        no_mutation: float = 0.4,
        arch_mutation: float = 0.2,
        new_layer: float = 0.2,
        param_mutation: float = 0.2,
        active_mutation: float = 0,
        hp_mutation: float = 0.2,
        hp_mutation_selection: list = ["lr", "batch_size"],
        mutation_strength: float = 0.1,
        evolution_steps: int = 10000,
        save_elite: bool = False,
        elite_path: str = "elite",
        tourn_size: int = 2,
        elitism: bool = True,
        hidden_size: list = [800, 600],
    ) -> None:
        agent_config = AgentConfig(
            max_timestep,
            memory_size,
            gamma,
            n_step,
            alpha,
            beta,
            tau,
            prior_eps,
            num_atoms,
            v_min,
            v_max,
            epsilon_start,
            epsilon_end,
            epsilon_decay,
        )

        trainer_config = TrainerConfig(
            batch_size,
            lr,
            seed,
            num_envs,
            device,
            learn_step,
            target_score,
            max_steps,
            evaluation_steps,
            evaluation_loop,
            learning_delay,
            n_step_memory,
            checkpoint,
            checkpoint_path,
            overwrite_checkpoints,
            use_wandb,
            wandb_api_key,
            accelerator,
        )

        hpo_config = HPOConfig(
            use_mutation,
            freq_evolution,
            population_size,
            no_mutation,
            arch_mutation,
            new_layer,
            param_mutation,
            active_mutation,
            hp_mutation,
            hp_mutation_selection,
            mutation_strength,
            evolution_steps,
            save_elite,
            elite_path,
            tourn_size,
            elitism,
        )

        network_config = NetworkConfig(hidden_size)

        training(
            agent_config,
            trainer_config,
            hpo_config,
            network_config,
            self.robot_config,
            self.sensor_config,
            self.env_config,
            self.render_config,
            self.pretrained_model,
        )

        return None

    def run(self):

        inference(
            self.robot_config,
            self.sensor_config,
            self.env_config,
            self.render_config,
            self.pretrained_model,
        )



import copy
import random

import numpy as np


class OUNoise:

    def __init__(self, size, seed, mu=0.0, theta=0.15, sigma=0.2):
        self.mu = mu * np.ones(size)
        self.theta = theta
        self.sigma = sigma
        self.seed = random.seed(seed)
        self.reset()

    def reset(self):
        self.state = copy.copy(self.mu)

    def sample(self):
        x = self.state
        dx = self.theta * (self.mu - x) + self.sigma * np.array(
            [random.random() for i in range(len(x))]
        )  # atualizar o estado interno
        self.state = x + dx
        return self.state



from dataclasses import dataclass

import numpy as np
import pymunk

from rnl.configs.config import RobotConfig


@dataclass
class Robot:
    """
    A class representing a robot with physical and sensor properties.
    """

    def __init__(self, robot_config: RobotConfig):
        """
        Initialize additional attributes after dataclass initialization.
        """
        # Constant acceleration due to gravity
        g = 9.81  # m/s²

        # Calculating the mass of the robot
        self.mass = robot_config.weight / g  # kg
        self.robot_radius = robot_config.base_radius
        self.wheel_base = robot_config.wheel_distance

        # Calculating the moment of inertia of the robot
        # I = (1/2) * m * r²
        self.inertia = 0.5 * self.mass * self.robot_radius**2
        self.body = pymunk.Body(self.mass, self.inertia)

    @staticmethod
    def create_space() -> pymunk.Space:
        """
        Create and return a new pymunk space with no gravity.
        """
        space = pymunk.Space()
        space.gravity = (0.0, 0.0)
        return space

    def create_robot(
        self,
        space: pymunk.Space,
        friction: float = 0.4,
        damping: float = 0.1,
        position_x: float = 0.0,
        position_y: float = 0.0,
    ) -> pymunk.Body:
        """
        Create and add the robot to the given pymunk space.
        """
        body = pymunk.Body(self.mass, self.inertia)
        body.position = (position_x, position_y)
        shape = pymunk.Circle(body, self.robot_radius)
        shape.friction = friction
        shape.damping = damping
        space.add(body, shape)
        return body

    def move_robot(
        self,
        space: pymunk.Space,
        robot_body: pymunk.Body,
        v_linear: float,
        v_angular: float,
    ) -> None:
        """
        Move the robot in the space with given linear and angular velocities.
        """
        direction = pymunk.Vec2d(np.cos(robot_body.angle), np.sin(robot_body.angle))
        robot_body.velocity = v_linear * direction
        robot_body.angular_velocity = v_angular
        space.step(1 / 60)

    def reset_robot(self, robot_body: pymunk.Body, x: float, y: float) -> None:
        """
        Reset the robot's position and velocity.
        """
        robot_body.position = (x, y)
        robot_body.angle = 0
        robot_body.velocity = (0, 0)
        robot_body.angular_velocity = 0



from dataclasses import dataclass, field
from typing import List


@dataclass
class EnvConfig:
    map_file: str = "rnl/data/map/map.pgm"
    random_mode: str = "normal"
    timestep: int = 1000
    grid_dimension: int = 5
    friction: float = 0.4
    porcentage_obstacles: float = 0.1
    max_step: int = 1000


@dataclass
class RenderConfig:
    fps: int = 1
    controller: bool = False
    rgb_array: bool = False


@dataclass
class SensorConfig:
    fov: float = 12.56637
    num_rays: int = 20
    min_range: float = 6.0
    max_range: float = 1.0


@dataclass
class RobotConfig:
    base_radius: float = 0.033
    vel_linear: List[float] = field(default_factory=lambda: [0.0, 1.0])
    vel_angular: List[float] = field(default_factory=lambda: [-1.0, 1.0])
    wheel_distance: float = 0.16
    weight: float = 1.0
    threshold: float = 0.01


@dataclass
class NetworkConfig:
    hidden_size: List[int] = field(default_factory=lambda: [800, 600])


@dataclass
class AgentConfig:
    max_timestep: int = 800000
    memory_size: int = 1000000
    gamma: float = 0.99
    n_step: int = 3
    alpha: float = 0.6
    beta: float = 0.4
    tau: float = 0.001
    prior_eps: float = 0.000001
    num_atoms: int = 51
    v_min: int = -200
    v_max: int = 200
    epsilon_start: float = 1.0
    epsilon_end: float = 0.1
    epsilon_decay: float = 0.995


@dataclass
class TrainerConfig:
    batch_size: int = 64
    lr: float = 0.0001
    seed: int = 1
    num_envs: int = 1
    device: str = "mps"
    learn_step: int = 10
    target_score: float = 200.0
    max_steps: int = 1000000
    evaluation_steps: int = 10000
    evaluation_loop: int = 1
    learning_delay: int = 0
    n_step_memory: int = 1
    checkpoint: int = 100
    checkpoint_path: str = "checkpoints"
    overwrite_checkpoints: bool = False
    use_wandb: bool = False
    wandb_api_key: str = ""
    accelerator: bool = False


@dataclass
class HPOConfig:
    use_mutation: bool = True
    freq_evolution: int = 10000
    population_size: int = 6
    no_mutation: float = 0.4
    arch_mutation: float = 0.2
    new_layer: float = 0.2
    param_mutation: float = 0.2
    active_mutation: float = 0
    hp_mutation: float = 0.2
    hp_mutation_selection: List[str] = field(
        default_factory=lambda: ["lr", "batch_size"]
    )
    mutation_strength: float = 0.1
    evolution_steps: int = 10000
    save_elite: bool = False
    elite_path: str = "elite"
    tourn_size: int = 2
    elitism: bool = True


class RandomizationDomainConfig:
    weight: List[float] = field(default_factory=lambda: [2, 500])
    base_radius: List[float] = field(default_factory=lambda: [0.15, 0.75])
    wheel_distance: List[float] = field(default_factory=lambda: [0.1, 1.5])
    threshold: List[float] = field(default_factory=lambda: [0.001, 0.1])
    fov: List[float] = field(default_factory=lambda: [3.14159, 12.56637])
    num_rays: List[int] = field(default_factory=lambda: [5, 60])
    range: List[float] = field(default_factory=lambda: [0.1, 40])
    timestep: List[int] = field(default_factory=lambda: [100, 10000])
    grid_dimension: List[int] = field(default_factory=lambda: [3, 500])
    friction: List[float] = field(default_factory=lambda: [0.1, 1.0])
    porcentage_obstacles: List[float] = field(default_factory=lambda: [0.01, 0.5])



import gymnasium as gym
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
from gymnasium import spaces
from mpl_toolkits.mplot3d import Axes3D, art3d

# from rnl.algorithms.rainbow import RainbowDQN
from rnl.configs.config import EnvConfig, RenderConfig, RobotConfig, SensorConfig
from rnl.engine.collision import Collision
from rnl.engine.utils import min_laser  # normalize
from rnl.engine.utils import angle_to_goal, distance_to_goal, get_reward
from rnl.environment.generate_world import Generator
from rnl.environment.robot import Robot
from rnl.environment.sensor import SensorRobot


class NaviEnv(gym.Env):
    def __init__(
        self,
        robot_config: RobotConfig,
        sensor_config: SensorConfig,
        env_config: EnvConfig,
        render_config: RenderConfig,
    ):
        super().__init__()
        state_size = sensor_config.num_rays + 4  # (action, distance, angle, reward)
        self.action_space = spaces.Discrete(6)  # action
        self.observation_space = spaces.Box(
            low=-np.inf, high=np.inf, shape=(state_size,), dtype=np.float32
        )

        self.generator = Generator(
            env_config.grid_dimension,
            env_config.porcentage_obstacles,
            env_config.random_mode,
        )
        self.random_state = env_config.random_mode
        self.collision = Collision()
        self.robot = Robot(robot_config)
        self.sensor = SensorRobot(sensor_config)

        self.space = self.robot.create_space()
        self.body = self.robot.create_robot(self.space)

        self.last_states = np.zeros(state_size)

        self.rgb_array = render_config.rgb_array
        self.timestep = 0
        self.max_timestep = env_config.max_step
        self.step_anim = env_config.max_step
        self.fps = render_config.fps
        self.threshold = robot_config.threshold
        self.grid_lenght = env_config.grid_dimension
        self.xmax = env_config.grid_dimension - 0.25
        self.ymax = env_config.grid_dimension - 0.25
        self.dist_max = np.sqrt(self.xmax**2 + self.ymax**2)

        self.segments = []
        # TODO
        self.controller = render_config.controller
        self.cumulated_reward = 0.0

        self.target_x = 0
        self.target_y = 0
        self.last_position_x = 0
        self.last_position_y = 0
        self.last_theta = 0
        self.last_measurement = 0
        self.vl = 0.01
        self.vr = 0.01
        self.init_distance = 0

        self.action = 0
        self.scalar = 10

        self.lidar_angle = np.linspace(0, 2 * np.pi, 20)
        self.measurement = np.zeros(20)

        # self.rainbow = RainbowDQN.load(
        #     "/microvault/rnl/checkpoints/model.pt",
        #     device="mps",
        # )

        if self.rgb_array:
            self.fig, self.ax = plt.subplots(1, 1, figsize=(6, 6))
            self.ax.remove()
            self.ax = self.fig.add_subplot(1, 1, 1, projection="3d")

            self.target = self.ax.plot3D(
                np.random.uniform(0, self.xmax),
                np.random.uniform(0, self.ymax),
                0,
                marker="x",
                markersize=3.0,
                color="red",
            )[0]

            self.agents = self.ax.plot3D(
                np.random.uniform(0, self.xmax),
                np.random.uniform(0, self.ymax),
                0,
                marker="o",
                markersize=1.0,
                color="orange",
            )[0]

            self.ani = animation.FuncAnimation

            self._init_animation(self.ax)
            if self.controller:
                self.fig.canvas.mpl_connect("key_press_event", self.on_key_press)

        self.reset()

    def on_key_press(self, event):
        if event.key == "up":
            self.action = 0
            self.vl = 0.05 * self.scalar
            self.vr = 0.0
        elif event.key == "down":
            self.action = 1
            self.vl = 0.1 * self.scalar
            self.vr = 0.0
        elif event.key == "left":
            self.action = 2
            self.vl = 0.05 * self.scalar
            self.vr = 0.15 * self.scalar
        elif event.key == "right":
            self.action = 3
            self.vl = 0.05 * self.scalar
            self.vr = -0.15 * self.scalar
        elif event.key == "w":
            self.action = 4
            self.vl = 0.01 * self.scalar
            self.vr = 0.0
        elif event.key == "d":
            self.action = 5
            self.vl = 0.05 * self.scalar
            self.vr = 0.3 * self.scalar
        elif event.key == "a":
            self.action = 6
            self.vl = 0.05 * self.scalar
            self.vr = -0.3 * self.scalar
        elif event.key == " ":
            self.vl = 0.0
            self.vr = 0.0

    def step_animation(self, i):
        action = 0  # self.rainbow.get_action(self.last_states, action_mask=None, training=True)[0]
        if not self.controller:

            if action == 0:
                self.vl = 0.05 * self.scalar
                self.vr = 0.0
            elif action == 1:
                self.vl = 0.1 * self.scalar
                self.vr = 0.0
            elif action == 2:
                self.vl = 0.05 * self.scalar
                self.vr = 0.15 * self.scalar
            elif action == 3:
                self.vl = 0.05 * self.scalar
                self.vr = -0.15 * self.scalar
            elif action == 4:
                self.vl = 0.01 * self.scalar
                self.vr = 0.0
            elif action == 5:
                self.vl = 0.05 * self.scalar
                self.vr = 0.3 * self.scalar
            elif action == 6:
                self.vl = 0.05 * self.scalar
                self.vr = -0.3 * self.scalar

        self.robot.move_robot(self.space, self.body, self.vl, self.vr)
        x, y = self.body.position.x, self.body.position.y

        intersections, lidar_measurements = self.sensor.sensor(
            x=x, y=y, segments=self.segments
        )

        dist = distance_to_goal(x, y, self.target_x, self.target_y)

        alpha = angle_to_goal(
            self.robot.body.position.x,
            self.robot.body.position.y,
            self.robot.body.position.angle,
            self.target_x,
            self.target_y,
        )

        diff_to_init = self.init_distance - dist

        collision, laser = min_laser(lidar_measurements, self.threshold)
        reward, done = get_reward(
            lidar_measurements, dist, collision, alpha, diff_to_init
        )

        states = np.concatenate(
            (
                np.array(lidar_measurements, dtype=np.float32),
                np.array([action], dtype=np.int16),
                np.array([dist], dtype=np.float32),
                np.array([alpha], dtype=np.float32),
                np.array([reward], dtype=np.float32),
            )
        )

        # TODO: normalize states

        # print(
        #     "\rReward: {:.2f}\tC. reward: {:.2f}\tDistance: {:.2f}\tAngle: {:.2f}\tAction: {:.2f}\tMin lidar: {:.2f}".format(
        #         states[23],
        #         self.cumulated_reward,
        #         states[21],
        #         states[22],
        #         states[20],
        #         np.min(states[:20]),
        #     ),
        # )

        self.last_states = states

        self._plot_anim(
            i,
            intersections,
            x,
            y,
            self.target_x,
            self.target_y,
        )

        truncated = self.timestep >= self.max_timestep

        if done or truncated:
            self._stop()

    def step(self, action):

        if action == 0:
            self.vl = 0.05 * self.scalar
            self.vr = 0.0
        elif action == 1:
            self.vl = 0.1 * self.scalar
            self.vr = 0.0
        elif action == 2:
            self.vl = 0.05 * self.scalar
            self.vr = 0.15 * self.scalar
        elif action == 3:
            self.vl = 0.05 * self.scalar
            self.vr = -0.15 * self.scalar
        elif action == 4:
            self.vl = 0.01 * self.scalar
            self.vr = 0.0
        elif action == 5:
            self.vl = 0.05 * self.scalar
            self.vr = 0.3 * self.scalar
        elif action == 6:
            self.vl = 0.05 * self.scalar
            self.vr = -0.3 * self.scalar

        self.robot.move_robot(self.space, self.body, self.vl, self.vr)
        x, y = self.body.position.x, self.body.position.y

        intersections, lidar_measurements = self.sensor.sensor(
            x=x, y=y, segments=self.segments
        )

        dist = distance_to_goal(x, y, self.target_x, self.target_y)

        alpha = angle_to_goal(
            self.robot.body.position.x,
            self.robot.body.position.y,
            self.robot.body.position.angle,
            self.target_x,
            self.target_y,
        )

        diff_to_init = self.init_distance - dist

        collision, laser = min_laser(lidar_measurements, self.threshold)
        reward, done = get_reward(
            lidar_measurements, dist, collision, alpha, diff_to_init
        )

        states = np.concatenate(
            (
                np.array(lidar_measurements, dtype=np.float32),
                np.array([action], dtype=np.int16),
                np.array([dist], dtype=np.float32),
                np.array([alpha], dtype=np.float32),
                np.array([reward], dtype=np.float32),
            )
        )

        self.last_states = states
        self.last_measurement = lidar_measurements

        self.cumulated_reward += reward
        self.timestep += 1

        truncated = self.timestep >= self.max_timestep

        return states, reward, done, truncated, {}

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)

        self.timestep = 0

        self.vr = 0.05
        self.vl = 0.0

        self.cumulated_reward = 0.0

        new_map_path, exterior, interior, all_seg = self.generator.world()
        self.segments = all_seg

        if self.rgb_array:
            for patch in self.ax.patches:
                patch.remove()

            self.ax.add_patch(new_map_path)
            art3d.pathpatch_2d_to_3d(new_map_path, z=0, zdir="z")

        while True:
            self.target_x = np.random.uniform(0, self.xmax)
            self.target_y = np.random.uniform(0, self.ymax)
            x = np.random.uniform(0, self.xmax)
            y = np.random.uniform(0, self.ymax)
            # theta = np.random.uniform(0, 2 * np.pi)

            if self.collision.check_collision(
                exterior, interior, x, y
            ) and self.collision.check_collision(
                exterior, interior, self.target_x, self.target_y
            ):
                break

        self.robot.reset_robot(self.body, x, y)
        intersections, measurement = self.sensor.sensor(
            x=self.robot.body.position.x, y=self.robot.body.position.y, segments=all_seg
        )

        self.last_measurement = measurement

        if self.rgb_array:
            self._plot_anim(
                0,
                intersections,
                self.robot.body.position.x,
                self.robot.body.position.y,
                self.target_x,
                self.target_y,
            )

        dist = distance_to_goal(
            self.robot.body.position.x,
            self.robot.body.position.y,
            self.target_x,
            self.target_y,
        )

        self.init_distance = dist

        alpha = angle_to_goal(
            self.robot.body.position.x,
            self.robot.body.position.y,
            self.robot.body.position.angle,
            self.target_x,
            self.target_y,
        )

        self.last_states = np.concatenate(
            (
                np.array(measurement, dtype=np.float32),  # measurement
                np.array([0], dtype=np.float32),  # velocity
                np.array([dist], dtype=np.float32),  # distance
                np.array([alpha], dtype=np.float32),  # angle
                np.array([0], dtype=np.float32),  # reward
            )
        )

        info = {}
        return self.last_states, info

    def render(self):
        self.ani = animation.FuncAnimation(
            self.fig,
            self.step_animation,
            init_func=self.reset,
            blit=False,
            frames=self.step_anim,
            interval=self.fps,
        )

        plt.show()

    def _stop(self):
        self.reset()
        self.ani.frame_seq = self.ani.new_frame_seq()
        self.step_anim = self.max_timestep

    def _init_animation(self, ax: Axes3D) -> None:
        """
        Initializes the 3D animation by setting up the environment and camera parameters.

        Parameters:
        ax (Axes3D): The 3D axes to be used for plotting.

        Returns:
        None
        """
        ax.set_xlim(0, self.grid_lenght)
        ax.set_ylim(0, self.grid_lenght)

        # ------ Create wordld ------ #

        path, _, _, _ = self.generator.world()

        ax.add_patch(path)

        art3d.pathpatch_2d_to_3d(path, z=0, zdir="z")

        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

        # # Hide grid lines
        ax.grid(False)

        # Hide axes ticks
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])

        # Hide axes
        ax.set_axis_off()

        # Set camera
        ax.elev = 20
        ax.azim = -155
        ax.dist = 1

        self.label = self.ax.text(
            0,
            0,
            0.05,
            self._get_label(0),
        )

        self.label.set_fontsize(14)
        self.label.set_fontweight("normal")
        self.label.set_color("#666666")

        self.fig.subplots_adjust(left=0, right=1, bottom=0.1, top=1)

    def get_reward_function(self):

        return self.reward_function

    @staticmethod
    def _get_label(timestep: int) -> str:
        """
        Generates a label for the environment.

        Parameters:
        timestep (int): The current time step.

        Returns:
        str: The generated label containing information about the environment and the current time step.
        """
        line1 = "Environment\n"
        line2 = "Time Step:".ljust(14) + f"{timestep:4.0f}\n"

        return line1 + line2

    def _plot_anim(
        self,
        i: int,
        intersections: np.ndarray,
        x: np.ndarray,
        y: np.ndarray,
        target_x: np.ndarray,
        target_y: np.ndarray,
    ) -> None:

        self.label.set_text(self._get_label(i))

        if hasattr(self, "laser_scatters"):
            for scatter in self.laser_scatters:
                scatter.remove()
            del self.laser_scatters

        self.laser_scatters = []
        for angle, intersection in zip(self.lidar_angle, intersections):
            if intersection is not None:
                scatter = plt.scatter(
                    intersection[0], intersection[1], color="g", s=0.5
                )
                self.laser_scatters.append(scatter)

        self.agents.set_data_3d(
            [x],
            [y],
            [0],
        )

        self.target.set_data_3d(
            [target_x],
            [target_y],
            [0],
        )



from dataclasses import dataclass
from typing import List, Tuple

import numpy as np

from rnl.configs.config import SensorConfig
from rnl.engine.collision import Collision


@dataclass
class SensorRobot:
    def __init__(self, sensor_config: SensorConfig):
        self.collision = Collision()
        self.max_range = sensor_config.max_range
        self.lidar_angle = np.linspace(0, sensor_config.fov, sensor_config.num_rays)

    def sensor(
        self, x: float, y: float, segments: List
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Perform sensor measurements based on the robot's position and environment segments.
        """
        seg = self.collision.filter_segments(segments, x, y, 6)
        intersections = self.collision.lidar_intersection(
            x, y, self.max_range, self.lidar_angle, seg
        )
        measurements = self.collision.lidar_measurement(
            x, y, self.max_range, self.lidar_angle, seg
        )
        return intersections, measurements



from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from shapely.geometry import LineString, Polygon
from skimage import measure

from rnl.engine.collision import Collision
from rnl.engine.world_generate import GenerateWorld


@dataclass
class Generator:
    def __init__(
        self,
        grid_lenght: int = 5,
        random: float = 0.1,  # 1300
        mode: str = "normal",
    ):
        self.grid_lenght = grid_lenght
        self.random = random
        self.collision = Collision()
        self.generate = GenerateWorld()

    @staticmethod
    def _map_border(m: np.ndarray) -> np.ndarray:
        """
        Adds a border around the given map array.

        Parameters:
        m (np.ndarray): The map array to add a border to.

        Returns:
        np.ndarray: The map array with a border added.
        """
        rows, columns = m.shape

        new = np.zeros((rows + 2, columns + 2))

        new[1:-1, 1:-1] = m

        return new

    @staticmethod
    def line_to_np_stack(line: LineString) -> np.ndarray:
        """
        Converts a LineString object to a numpy array stack of points.

        Parameters:
        line (LineString): The LineString object to convert.

        Returns:
        np.ndarray: The numpy array stack of points representing the LineString.
        """
        coords = np.array(line.coords)

        return np.vstack((coords[:, 0], coords[:, 1])).T

    # TODO
    def upscale_map(self, original_map, resolution):
        new_shape = (
            original_map.shape[0] * int(1 / resolution),
            original_map.shape[1] * int(1 / resolution),
        )

        new_map = np.zeros(new_shape)

        for i in range(new_shape[0]):
            for j in range(new_shape[1]):
                original_value = original_map[int(i * resolution), int(j * resolution)]
                new_map[i, j] = original_value

        return new_map

    def world(self) -> Tuple[PathPatch, List, List, List]:
        """
        Generates a maze world.

        Returns:
        Tuple[PathPatch, Polygon, List]: A tuple containing:
        - PathPatch: The PathPatch object representing the maze.
        - Polygon: The Polygon object representing the maze boundaries.
        - List: List of LineString segments representing the maze segments.
        """
        m = self.generate.generate_maze(
            map_size=self.grid_lenght,
            decimation=0.0,
            min_blocks=5,
            num_cells_togo=1300,
        )

        border = self._map_border(m)
        map_grid = 1 - border

        contours = measure.find_contours(map_grid, 0.5)

        height, width = map_grid.shape
        exterior = []

        """
        #---------1---------#
        |                   |
        |                   |
        4                   2
        |                   |
        |                   |
        #---------3---------#
        """

        # 1
        for x in range(width):
            exterior.append((x, height - 1))

        # 2
        for y in range(height - 2, -1, -1):
            exterior.append((width - 1, y))

        # 3
        for x in range(width - 2, -1, -1):
            exterior.append((x, 0))

        # 4
        for y in range(1, height - 1):
            exterior.append((0, y))

        interiors = []
        segments = []

        for n, contour in enumerate(contours):
            poly = []
            for idx, vertex in enumerate(contour):
                poly.append((vertex[1], vertex[0]))

            interiors.append(poly)

            interior_segment = LineString(poly)
            segments.append(interior_segment)

        exterior_segment = LineString(exterior + [exterior[0]])
        segments.insert(0, exterior_segment)

        ext_segments = self.convert_to_segments(exterior)
        int_segments = [self.convert_to_segments(interior) for interior in interiors]

        stacks = [self.line_to_np_stack(line) for line in segments]

        segment = self.collision.extract_seg_from_polygon(stacks)

        poly = Polygon(exterior, holes=interiors).buffer(0)

        path = Path.make_compound_path(
            Path(np.asarray(poly.exterior.coords)[:, :2]),
            *[Path(np.asarray(ring.coords)[:, :2]) for ring in poly.interiors]
        )

        path_patch = PathPatch(
            path, edgecolor=(0.1, 0.2, 0.5, 0.15), facecolor=(0.1, 0.2, 0.5, 0.15)
        )

        return path_patch, ext_segments, int_segments, segment

    @staticmethod
    def convert_to_segments(polygon):
        segments = []
        n = len(polygon)
        for i in range(n):
            x1, y1 = polygon[i]
            x2, y2 = polygon[(i + 1) % n]
            segments.append((x1, y1, x2, y2))
        return segments



import warnings
from dataclasses import dataclass, field

import torch


@dataclass
class SpecItem:
    tensor: torch.Tensor
    tensor_name: str
    module_name: str = None
    changing: bool = None
    check_nan: bool = False
    check_inf: bool = False
    _old_copy: torch.Tensor = field(init=False, default=None)

    def __post_init__(self):
        if self.changing is not None:
            self._old_copy = self.tensor.detach().clone()

    @property
    def name(self):
        if self.module_name is None:
            return self.tensor_name
        else:
            return f"Module {self.module_name}'s {self.tensor_name}"

    def update(
        self,
        tensor_name,
        module_name=None,
        changing=None,
        check_nan=False,
        check_inf=False,
    ):
        if (tensor_name != self.tensor_name) or (
            module_name is not None and module_name != self.module_name
        ):
            old_name = self.name
            self.tensor_name = tensor_name
            if module_name is not None:
                self.module_name = module_name
            warnings.warn(f"{old_name} is renamed as {self.name}")
        if changing is not None:
            self.changing = changing
        if check_nan:
            self.check_nan = True
        if check_inf:
            self.check_inf = True

    def validate(self):
        error_items = []
        if self.changing is not None:
            error_items.append(self.validate_changing())
        if self.check_nan:
            error_items.append(self.validate_nan())
        if self.check_inf:
            error_items.append(self.validate_inf())

        error_items = [_ for _ in error_items if _ is not None]

    def validate_changing(self):
        if self.changing:
            if torch.equal(self.tensor, self._old_copy):
                return f"{self.name} should change."
        else:
            if not torch.equal(self.tensor, self._old_copy):
                return f"{self.name} should not change."

        self._old_copy = self.tensor.detach().clone()

    def validate_nan(self):
        if torch.any(torch.isnan(self.tensor)).item():
            return f"{self.name} contains NaN."

    def validate_inf(self):
        if torch.any(torch.isinf(self.tensor)).item():
            return f"{self.name} contains inf."


@dataclass
class ParamSpec:
    specs: dict = field(default_factory=dict)

    def add(
        self,
        tensor,
        tensor_name,
        module_name=None,
        changing=None,
        check_nan=False,
        check_inf=False,
    ):
        if tensor in self.specs:
            self.specs[tensor].update(
                tensor_name=tensor_name,
                module_name=module_name,
                changing=changing,
                check_nan=check_nan,
                check_inf=check_inf,
            )
        else:
            self.specs[tensor] = SpecItem(
                tensor=tensor,
                tensor_name=tensor_name,
                module_name=module_name,
                changing=changing,
                check_nan=check_nan,
                check_inf=check_inf,
            )

    def validate(self):
        error_strings = []
        for spec in self.specs.values():
            error_string = spec.validate()
            if len(error_string) > 0:
                error_strings.append(error_string)
        if len(error_strings) > 0:
            error_msg = "\n".join(error_strings)
            raise RuntimeError(
                f"The following errors are detected while training:\n{error_msg}"
            )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class NoisyLinear(nn.Module):
    """The Noisy Linear Neural Network class.

    :param in_features: Input features size
    :type in_features: int
    :param out_features: Output features size
    :type out_features: int
    :param std_init: Standard deviation, defaults to 0.5
    :type std_init: float, optional
    """

    def __init__(self, in_features, out_features, std_init=0.5):
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.std_init = std_init

        self.weight_mu = nn.Parameter(torch.FloatTensor(out_features, in_features))
        self.weight_sigma = nn.Parameter(torch.FloatTensor(out_features, in_features))
        self.register_buffer(
            "weight_epsilon", torch.FloatTensor(out_features, in_features)
        )

        self.bias_mu = nn.Parameter(torch.FloatTensor(out_features))
        self.bias_sigma = nn.Parameter(torch.FloatTensor(out_features))
        self.register_buffer("bias_epsilon", torch.FloatTensor(out_features))

        self.reset_parameters()
        self.reset_noise()

    def forward(self, x):
        """Returns output of neural network.

        :param x: Neural network input
        :type x: torch.Tensor()
        """
        weight_epsilon = self.weight_epsilon.to(x.device)
        bias_epsilon = self.bias_epsilon.to(x.device)

        if self.training:
            weight = self.weight_mu + self.weight_sigma.mul(weight_epsilon)
            bias = self.bias_mu + self.bias_sigma.mul(bias_epsilon)
        else:
            weight = self.weight_mu
            bias = self.bias_mu

        return F.linear(x, weight, bias)

    def reset_parameters(self):
        """Resets neural network parameters."""
        mu_range = 1 / math.sqrt(self.in_features)

        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.weight_sigma.data.fill_(self.std_init / math.sqrt(self.in_features))

        self.bias_mu.data.uniform_(-mu_range, mu_range)
        self.bias_sigma.data.fill_(self.std_init / math.sqrt(self.out_features))

    def reset_noise(self):
        """Resets neural network noise."""
        epsilon_in = self._scale_noise(self.in_features)
        epsilon_out = self._scale_noise(self.out_features)

        self.weight_epsilon.copy_(epsilon_out.ger(epsilon_in))
        self.bias_epsilon.copy_(epsilon_out)

    def _scale_noise(self, size):
        """Returns noisy tensor.

        :param size: Tensor of same size as noisy output
        :type size: torch.Tensor()
        """
        x = torch.randn(size)
        return x.sign().mul_(x.abs().sqrt_())



from rnl.components.replay_buffer import MultiStepReplayBuffer, PrioritizedReplayBuffer
from rnl.configs.config import (
    AgentConfig,
    EnvConfig,
    HPOConfig,
    NetworkConfig,
    RenderConfig,
    RobotConfig,
    SensorConfig,
    TrainerConfig,
)
from rnl.environment.environment_navigation import NaviEnv
from rnl.hpo.mutation import Mutations
from rnl.hpo.tournament import TournamentSelection
from rnl.training.train_off_policy import train_off_policy
from rnl.training.utils import create_population, make_vect_envs


def training(
    agent_config: AgentConfig,
    trainer_config: TrainerConfig,
    hpo_config: HPOConfig,
    network_config: NetworkConfig,
    robot_config: RobotConfig,
    sensor_config: SensorConfig,
    env_config: EnvConfig,
    render_config: RenderConfig,
    pretrained_model,
):

    env = make_vect_envs(
        env_name=NaviEnv,
        num_envs=trainer_config.num_envs,
        robot_config=robot_config,
        sensor_config=sensor_config,
        env_config=env_config,
        render_config=render_config,
    )

    net_config = {
        "arch": "mlp",
        "hidden_size": network_config.hidden_size,  # Network hidden size
    }

    state_dim = env.single_observation_space.shape
    action_dim = env.single_action_space.n

    field_names = ["state", "action", "reward", "next_state", "termination"]
    memory = PrioritizedReplayBuffer(
        memory_size=agent_config.memory_size,
        field_names=field_names,
        num_envs=trainer_config.num_envs,
        alpha=agent_config.alpha,
        gamma=agent_config.gamma,
        device=trainer_config.device,
    )

    n_step_memory = MultiStepReplayBuffer(
        memory_size=agent_config.memory_size,
        field_names=field_names,
        num_envs=trainer_config.num_envs,
        n_step=agent_config.n_step,
        gamma=agent_config.gamma,
        device=trainer_config.device,
    )

    agent_pop = create_population(
        algo="RainbowDQN",  # Algorithm
        state_dim=state_dim,  # State dimension
        action_dim=action_dim,  # Action dimension
        one_hot=False,  # One-hot encoding
        net_config=net_config,  # Network configuration
        agent_config=agent_config,  # Agent configuration
        trainer_config=trainer_config,  # Trainer configuration
        population_size=hpo_config.population_size,  # Population size
        num_envs=trainer_config.num_envs,  # Number of vectorized environments
        device=trainer_config.device,
    )

    tournament = TournamentSelection(
        tournament_size=hpo_config.tourn_size,  # Tournament selection size
        elitism=hpo_config.elitism,  # Elitism in tournament selection
        population_size=hpo_config.population_size,  # Population size
        eval_loop=trainer_config.evaluation_loop,  # Evaluate using last N fitness scores
    )

    mutations = Mutations(
        algo="Rainbow DQN",  # Algorithm
        no_mutation=hpo_config.no_mutation,  # No mutation
        architecture=hpo_config.arch_mutation,  # Architecture mutation
        new_layer_prob=hpo_config.new_layer,  # New layer mutation
        parameters=hpo_config.param_mutation,  # Network parameters mutation
        activation=hpo_config.active_mutation,  # Activation layer mutation
        rl_hp=hpo_config.hp_mutation,  # Learning HP mutation
        rl_hp_selection=hpo_config.hp_mutation_selection,  # Learning HPs to choose from
        mutation_sd=hpo_config.mutation_strength,  # Mutation strength
        arch="mlp",  # Network architecture
        rand_seed=trainer_config.seed,  # Random seed
        device=trainer_config.device,
    )

    config_dict = {
        "Agent Config": agent_config.__dict__,
        "Trainer Config": trainer_config.__dict__,
        "HPO Config": hpo_config.__dict__,
        "Network Config": network_config.__dict__,
    }

    for config_name, config_values in config_dict.items():
        print(f"\n#------ {config_name} ----#")
        max_key_length = max(len(key) for key in config_values.keys())
        for key, value in config_values.items():
            print(f"{key.ljust(max_key_length)} : {value}")

    trained_pop, pop_fitnesses = train_off_policy(
        env=env,
        env_name="NaviEnv-v0",
        algo="Rainbow DQN",
        pop=agent_pop,
        memory=memory,
        n_step_memory=n_step_memory,
        max_steps=trainer_config.max_steps,
        evo_steps=hpo_config.evolution_steps,
        eval_steps=trainer_config.evaluation_steps,
        eval_loop=trainer_config.evaluation_loop,
        learning_delay=trainer_config.learning_delay,
        target=trainer_config.target_score,
        n_step=True,
        per=True,
        tournament=tournament,
        mutation=mutations,
        wb=False,
        checkpoint=100,
        checkpoint_path="RainbowDQN.pt",
    )


def inference(
    robot_config: RobotConfig,
    sensor_config: SensorConfig,
    env_config: EnvConfig,
    render_config: RenderConfig,
    pretrained_model=False,
):

    text = [
        r"+--------------------+",
        r" ____  _   _ _",
        r"|  _ \| \ | | |",
        r"| |_) |  \| | |",
        r"|  _ <| |\  | |___",
        r"|_| \_\_| \_|_____|",
        r"_____________________",
    ]

    for line in text:
        print(line)

    env = NaviEnv(robot_config, sensor_config, env_config, render_config)

    env.reset()
    env.render()



from gymnasium.envs.registration import register

register(
    id="NaviEnv-v0",
    entry_point="rnl.environment.environment_navigation:NaviEnv",  # Adjust this if NaviEnv is in a different module
    max_episode_steps=500000,
)
